library(stringr)
library(StereoMorph)
library(geomorph)
# Some custom functions of mine
source("utility_functions.R")

# Load plastron data from external tps file
plastron <- readland.tps("raw_landmarks/plastron.tps", specID = "ID")
# Load specimen details
specimens <- read.csv("specimens.csv", header = T, as.is = T)[, 5:10]

nb_landmarks <- dim(plastron)[1]
nb_replicates <- dim(plastron)[3]

# Load landmark definitions, which specify the laterality of each landmark.
landmark_definitions <- read.table("landmark_definitions_plas.txt", header = TRUE)

ant_lobe <- plastron[which(landmark_definitions$lobe == "anterior"), , ]
ant_lobe_side <- na.omit(landmark_definitions$side[landmark_definitions$lobe == "anterior"])
ant_lobe_side <- as.character(ant_lobe_side)
ant_lobe_names <- vector(length = length(ant_lobe_side))
ant_lobe_names[ant_lobe_side == "left"] <- paste0(1:sum(ant_lobe_side == "left"), "_L")
ant_lobe_names[ant_lobe_side == "right"] <- paste0(1:sum(ant_lobe_side == "right"), "_R")
ant_lobe_names[ant_lobe_side == "medial"] <- paste0(1:sum(ant_lobe_side == "medial"), "_M")

pos_lobe <- plastron[which(landmark_definitions$lobe == "posterior"), , ]
pos_lobe_side <- na.omit(landmark_definitions$side[landmark_definitions$lobe == "posterior"])
pos_lobe_side <- as.character(pos_lobe_side)
pos_lobe_names <- vector(length = length(pos_lobe_side))
pos_lobe_names[pos_lobe_side == "left"] <- paste0(1:sum(pos_lobe_side == "left"), "_L")
pos_lobe_names[pos_lobe_side == "right"] <- paste0(1:sum(pos_lobe_side == "right"), "_R")
pos_lobe_names[pos_lobe_side == "medial"] <- paste0(1:sum(pos_lobe_side == "medial"), "_M")

# Processing of the anterior lobe ####

for (i in 1:nb_replicates) {
  M <- ant_lobe[, , i]
  rownames(M) <- ant_lobe_names
  M2 <- reflectMissingLandmarks(M, average=TRUE)
  ant_lobe[, , i] <- M2$lm.matrix[match(rownames(M), rownames(M2$lm.matrix)), ]
}

# Excluding specimens with any NA landmarks
ant_lobe <- exclude_specimen(ant_lobe)
# Calculate the average between replicates of specimens
ant_lobe <- average_replicates(ant_lobe)

# Fetching the "species" names ("lisu": least inclusive systematic unit, can also be subspecies) and maturity data by matching the call_code (specimen voucher numbers)
specimen_match <- match(dimnames(ant_lobe)[[3]], specimens$call_code)
ant_lobe <- ant_lobe[, , -which(is.na(specimen_match))]
specimen_match <- na.omit(specimen_match)
dimnames(ant_lobe)[[3]] <- str_extract(specimens$lisu[specimen_match], "(\\S+)(\\s)(\\S+)") # This will merge all subspecies into their parent species.
maturity <- specimens$maturity[specimen_match]

# Exclude juveniles
maturity[which(maturity=="")] <- "adult" # Specimens without maturity data will be assumed to be adults
maturity[which(is.na(maturity))] <- "adult" # Specimens without maturity data will be assumed to be adults
ant_lobe <- keep_adults(ant_lobe, maturity)

# Calculate species shape by averaging specimens
ant_lobe <- average_species(ant_lobe)

# Discarding some unidentified and hybrid specimens
ant_lobe <- ant_lobe[, , - which(dimnames(ant_lobe)[[3]] == "Mauremys iversoni")] # Remove Mauremys iversoni because it is a hybrid between Cuora trifasciata and Mauremys mutica mutica
ant_lobe <- ant_lobe[, , - which(dimnames(ant_lobe)[[3]] == "Mauremys sp")] # Remove Mauremys sp
ant_lobe <- ant_lobe[, , - which(dimnames(ant_lobe)[[3]] == "Gopherus polyphemus")] # This species won't be used
# ant_lobe <- ant_lobe[, , - which(dimnames(ant_lobe)[[3]] == "Malaclemys terrapin")] # Remove Mauremys sp
ant_lobe <- ant_lobe[, , - which(dimnames(ant_lobe)[[3]] == "Graptemys geographica")] # This species won't be used
ant_lobe <- ant_lobe[, , - which(dimnames(ant_lobe)[[3]] == "Graptemys nigrinoda")] # This species won't be used
ant_lobe <- ant_lobe[, , - which(dimnames(ant_lobe)[[3]] == "Graptemys versa")] # This species won't be used


# Update the number of species
nb_species <- dim(ant_lobe)[3]

# A quick visual inspection to make sure the species names are correct.
plot(ant_lobe[, , which(dimnames(ant_lobe)[[3]] == "Sacalia bealei")], asp = 1)
# It looks OK!

# Load the landmark definitions table, which indicates which landmarks belong to each side. I am only keeping the medial and righ-side landmarks.

keep_lands_ant <- c(which(ant_lobe_side == "right"), which(ant_lobe_side == "medial"))

# Store unaligned configurations (only right side)
writeland.tps(ant_lobe[keep_lands_ant, , ], "superpositions/ant_lobe.tps")

# Procrustes superimpostiion
gpa_ant_lobe <- gpagen(ant_lobe)
writeland.tps(gpa_ant_lobe$coords[keep_lands_ant, , ], "superpositions/ant_lobe_gpa.tps")

# Processing of the posterior lobe ####

for (i in 1:nb_replicates) {
  M <- pos_lobe[, , i]
  rownames(M) <- pos_lobe_names
  M2 <- reflectMissingLandmarks(M, average=TRUE)
  pos_lobe[, , i] <- M2$lm.matrix[match(rownames(M), rownames(M2$lm.matrix)), ]
}

# Excluding specimens with any NA landmarks
pos_lobe <- exclude_specimen(pos_lobe)
# Calculate the average between replicates of specimens
pos_lobe <- average_replicates(pos_lobe)

# Fetching the "species" names ("lisu": least inclusive systematic unit, can also be subspecies) and maturity data by matching the call_code (specimen voucher numbers)
specimen_match <- match(dimnames(pos_lobe)[[3]], specimens$call_code)
pos_lobe <- pos_lobe[, , -which(is.na(specimen_match))]
specimen_match <- na.omit(specimen_match)
dimnames(pos_lobe)[[3]] <- str_extract(specimens$lisu[specimen_match], "(\\S+)(\\s)(\\S+)") # This will merge all subspecies into their parent species.
maturity <- specimens$maturity[specimen_match]

# Exclude juveniles
maturity[which(maturity=="")] <- "adult" # Specimens without maturity data will be assumed to be adults
maturity[which(is.na(maturity))] <- "adult" # Specimens without maturity data will be assumed to be adults
pos_lobe <- keep_adults(pos_lobe, maturity)

# Calculate species shape by averaging specimens
pos_lobe <- average_species(pos_lobe)

# Discarding some unidentified and hybrid specimens
pos_lobe <- pos_lobe[, , - which(dimnames(pos_lobe)[[3]] == "Mauremys iversoni")] # Remove Mauremys iversoni because it is a hybrid between Cuora trifasciata and Mauremys mutica mutica
pos_lobe <- pos_lobe[, , - which(dimnames(pos_lobe)[[3]] == "Mauremys sp")] # Remove Mauremys sp
pos_lobe <- pos_lobe[, , - which(dimnames(pos_lobe)[[3]] == "Gopherus polyphemus")] # This species won't be used
# pos_lobe <- pos_lobe[, , - which(dimnames(pos_lobe)[[3]] == "Malaclemys terrapin")] # Remove Mauremys sp
pos_lobe <- pos_lobe[, , - which(dimnames(pos_lobe)[[3]] == "Graptemys geographica")] # This species won't be used
pos_lobe <- pos_lobe[, , - which(dimnames(pos_lobe)[[3]] == "Graptemys nigrinoda")] # This species won't be used
pos_lobe <- pos_lobe[, , - which(dimnames(pos_lobe)[[3]] == "Graptemys versa")] # This species won't be used

# Update the number of species
nb_species <- dim(pos_lobe)[3]

# A quick visual inspection to make sure the species names are correct.
plot(pos_lobe[, , which(dimnames(pos_lobe)[[3]] == "Pangshura tentoria")], asp = 1)
# It looks OK!

# Load the landmark definitions table, which indicates which landmarks belong to each side. I am only keeping the medial and righ-side landmarks.

keep_lands_pos <- c(which(pos_lobe_side == "right"), which(pos_lobe_side == "medial"))

# Store unaligned configurations (only right side)
writeland.tps(pos_lobe[keep_lands_pos, , ], "superpositions/pos_lobe.tps")

# Procrustes superimpostiion
gpa_pos_lobe <- gpagen(pos_lobe)
writeland.tps(gpa_pos_lobe$coords[keep_lands_pos, ,], "superpositions/pos_lobe_gpa.tps")

# Generate TNT files ####
# writeland.tnt(list(ant_lobe = ant_lobe[keep_lands_ant, ,], pos_lobe = pos_lobe[keep_lands_pos, ,]), "tnt_data/plastron.tnt", log = NULL)
# writeland.tnt(ant_lobe[keep_lands_ant, ,], "tnt_data/ant_lobe.tnt", log = NULL)
# writeland.tnt(pos_lobe[keep_lands_pos, ,], "tnt_data/pos_lobe.tnt", log = NULL)

# writeland.tnt(list(ant_lobe = gpa_ant_lobe$coords[keep_lands_ant, ,], pos_lobe = gpa_pos_lobe$coords[keep_lands_pos, ,]), "tnt_data/plastron_gpa.tnt", log = NULL)
# writeland.tnt(gpa_ant_lobe$coords[keep_lands_ant, ,], "tnt_data/ant_lobe_gpa.tnt", log = NULL)
# writeland.tnt(gpa_pos_lobe$coords[keep_lands_pos, ,], "tnt_data/pos_lobe_gpa.tnt", log = NULL)
