# General array operations ####

# Split an array along a given dimension. Function provided by StackOverflow user flodel.
split_along_dim <- function(a, n) setNames(lapply(split(a, arrayInd(seq_along(a), dim(a))[, n]), array, dim = dim(a)[-n], dimnames(a)[-n]), dimnames(a)[[n]])

# Custom logical functions ####

'%notin%' <- Negate('%in%')

notna <- Negate(is.na)

# Specimen management ####

# Exclude specimens with any NA landmarks
exclude_specimen <- function (dat, criterion = "NA") {
  nb_specimens <- dim(dat)[[3]]
  exclude <- rep(F, nb_specimens)
  exclude <- sapply(c(1:nb_specimens), function (x) {
    if (any(is.na(dat[, , x]))) exclude[x] <- T
    else exclude[x] <- F
  })
  dat <- dat[, , - which(exclude)]
}

# Exclude juvenile specimens
keep_adults <- function (dat, maturity) {
  nb_specimens <- dim(dat)[[3]]
  exclude <- rep(F, nb_specimens)
  exclude <- sapply(c(1:nb_specimens), function (x) {
    if (maturity[x] == "juvenile" | maturity[x] == "subadult") exclude[x] <- T
    else exclude[x] <- F
  })
  dat <- dat[, , - which(exclude)]
}

# Superimpose and average replicates of specimens. Each replicate name is composed of the voucher number, a delimitation character ($ by default) and a postfix replicate identificator (normally a lowercase letter).
average_replicates <- function(A, delim = "\\$", postfix = letters) {
  # Average species of the same genus in array A with configuration names in the format "Genus_species".
  # Does not allow multiple specimens per species.
  require(geomorph)
  require(abind)
  
  # Create vector of species names
  tags <- dimnames(A)[[3]]
  # Create vector with the postfix patterns
  postfix_patterns <- sapply(postfix, function (x) paste(delim, x, sep=""))
  postfix_remove <- vector("numeric", length=0)
  for (i in 1:length(postfix_patterns)) {
    if (length(grep(postfix_patterns[i], tags) == 0)) postfix_remove <- c(postfix_remove, i)
  }
  postfix_patterns <- postfix_patterns[postfix_remove]
  # Remove postfix patterns from tags
  for (i in 1:length(postfix_patterns)) { # Could be replaced by sapply?
    tags <- sub(postfix_patterns[i], "", tags)
  }
  # Do generalised Procrustes alignments of the replicates
  # replicate_gpa <- lapply(tags, function(x) if (dim(A[, , tags == x])[3] > 1) gpagen(A[, , tags == x])$consensus)
  replicate_gpa <- vector("list", length(tags))
  for (i in 1:length(tags)) {
    print(i)
    if (length(dim(A[, , tags == tags[i]])) > 2) replicate_gpa[[i]] <- gpagen(A[, , tags == tags[i]])$consensus
  }
  replicate_gpa <- array(unlist(replicate_gpa), dim = c(nrow(replicate_gpa[[1]]), ncol(replicate_gpa[[1]]), length(replicate_gpa)))
  dimnames(replicate_gpa)[[3]] <- tags
  # Return the resulting array
  replicate_gpa
}

# Superimpose and average specimens by species (or lower analytical unit).
average_species <- function(dat) {
  species_names <- unique(dimnames(dat)[[3]])
  nb_species <- length(species_names)
  species_average <- array(NA, dim = c(dim(dat)[1], 3, nb_species))
  dimnames(species_average)[[3]] <- vector("list", nb_species)
  
  for (i in 1:nb_species) {
    conspecifics <- which(dimnames(dat)[[3]] == species_names[i])
    if (length(conspecifics) < 2) {
      species_average[, , i] <- dat[, , conspecifics]
      dimnames(species_average)[[3]][i] <- species_names[i]
      next
    }
    species_array <- dat[, , conspecifics]
    mean_species <- gpagen(species_array)$consensus
    species_average[, , i] <- mean_species
    dimnames(species_average)[[3]][i] <- species_names[i]
  }
  # dimnames(species_average)[[3]] <- species_names
  return(species_average)
}

# Export a continous character matrix to TNT format.
# Species names should be given as row names. Character names (optional) are given by column names.
# `dec` is used to round the values.
writecont.tnt <- function(mat, file, dec = 3) {
  out_file <- file(file, "w")
  n_char <- ncol(mat)
  n_tip <- nrow(mat)
  writeLines("NSTATES CONT ;", out_file)
  writeLines("XREAD", out_file)
  tnt_comment <- paste0("\'TNT file generated with the R function `writecont.tnt` on ", date(), ".\'")
  writeLines(tnt_comment, out_file)
  writeLines(paste(n_char, n_tip), out_file)
  writeLines("&[continuous]", out_file)
  tip_names <- rownames(mat)
  tip_names <- gsub(" ", "_", tip_names)
  rownames(mat) <- tip_names
  cnames <- colnames(mat)
  write.table(round(mat, dec), file = out_file, append = TRUE, col.names = FALSE, quote = FALSE)
  if (!is.null(cnames)) {
    writeLines(";\n\nCNAMES", out_file)
    for (i in 0:(length(cnames) - 1)) {
      writeLines(paste0("{  ", i, " ", cnames[i + 1], ";"), out_file)
    }
  }
  writeLines(";\nPROC /;\n", out_file)
  close.connection(out_file)
}

# Write 2D or 3D landmark data to a file in TNT format. `A` can be either an
# array with landmarks from a single configuration or a list of arrays of
# different configurations. Some species may not be included in all the arrays.
# If that happens, `allow_missing = FALSE` will drop those species from the
# entire dataset. `allow_missing = TRUE` will keep all the species and fill in
# with question marks as necessary. A file name to write out a TNT log can be
# given with the `log` parameter. writeland.tnt will also add an ECHO command.
# `dec` is used to round the values
writeland.tnt <- function (A, file, dec = 3, allow_missing = FALSE, log = NULL) {
  .writearray.tnt <- function (A, n, global_names, file) {
    # Write an array of landmark data in TNT format to an open file connection.
    A <- round(A, dec)
    k <- dim(A)[2] # Number of dimensions
    p <- dim(A)[1] # Number of landmarks
    dimnames(A)[[3]] <- gsub(" ", "_", dimnames(A)[[3]]) -> item_names
    
    if (k == 2) {
      writeLines("& [landmark 2D]", file)
    } else {
      writeLines("& [landmark 3D]", file)
    }
    for (i in 1:n) { # Iterate over terminal items
      this_line <- global_names[i]
      if (global_names[i] %in% item_names) {
        for (j in 1:p) { # Iterate over landmarks
          coords <- sapply(A[j, , global_names[i]], function (x) ifelse ((x >= 0), paste0("+", x), x))
          coords <- paste(coords, collapse = ",")
          coords <- gsub("NA,NA,NA", "?", coords) # This is for missing landmarks
          this_line <- paste(this_line, coords, collapse = "  ")
        }
      } else {
        this_line <- paste(this_line, paste(rep("?", p), collapse = " "))
      }
      writeLines(this_line, file)
    }
  }
  
  file <- file(file, "w")
  if (! is.null(log)) writeLines(paste0("LOG ", log, ";\nECHO =;"), file)
  writeLines("NSTATES CONT;", file)
  writeLines("XREAD", file)
  tnt_comment <- paste0("\'TNT file generated with the R function `writeland.tnt` on ", date(), ".\'")
  writeLines(tnt_comment, file)
  if (class(A) == "list") {
    item_names <- vector(mode = "character")
    if (allow_missing) {
      for (a in A) {
        item_names <- c(item_names, setdiff(dimnames(a)[[3]], item_names))
      }
    } else {
      item_names <- Reduce(intersect, lapply(A, function (x) dimnames(x)[[3]]))
    }
    item_names <- gsub(" ", "_", item_names)
    n <- length(item_names) # Number of terminal items (sp, ssp, individuals, &c)
    cnames <- names(A)
    writeLines(paste(length(A), n), file)
    for (a in A) {
      .writearray.tnt(a, n, item_names, file)
    }
  } else {
    item_names <- dimnames(A)[[3]]
    item_names <- gsub(" ", "_", item_names)
    n <- length(item_names)
    cnames <- NULL
    writeLines(paste("1", n), file)
    .writearray.tnt(A, n, item_names, file)
  }
  if (! is.null(cnames)) {
    writeLines(";\n\nCNAMES", file)
    for (i in 0:(length(cnames) - 1)) {
      writeLines(paste0("{  ", i, " ", cnames[i + 1], ";"), file)
    }
  }
  writeLines(";\n", file)
  if (! is.null(log)) writeLines("LOG /;", file)
  writeLines("PROC /;\n", file)
  close.connection(file)
}

# Write a phylo or multiPhylo object in the TNT parenthetical tree format. TNT
# cannot read trees without a matrix in memory, so this function offers the
# option to create a dummy matrix. Trees can have diffent tip labels.
write.tree.tnt <- function(tr, out_file, dummy = TRUE) {
  require(ape)
  out_file <- file(out_file, "w")
  if (class(tr) == "multiPhylo") {
    n_tr <- length(tr)
    tip_labels <- Reduce(union, lapply(tr, function(x)
      x$tip.label))
  } else {
    n_tr <- 1
    tip_labels <- tr$tip.label
    tr <- list(tr)
  }
  
  if (dummy) {
    writeLines("NSTATES 8 ;", out_file)
    writeLines("XREAD", out_file)
    writeLines(paste("1", length(tip_labels)), out_file)
    mat <- cbind(tip_labels, 0)
    write.table(
      mat,
      file = out_file,
      col.names = FALSE,
      row.names = FALSE,
      append = TRUE,
      quote = FALSE
    )
    writeLines(";", out_file)
  }
  
  tr <-
    lapply(tr, function (x) {
      x$edge.length <- NULL
      x$node.label <- NULL
      x
    })
  
  writeLines("TREAD", out_file)
  for (i in 1:n_tr) {
    nwkstr <- write.tree(tr[[i]])
    tntstr <- gsub(",", " ", nwkstr)
    tntstr <- gsub(";", "*", tntstr)
    writeLines(tntstr, out_file)
  }
  writeLines(";\nPROC/;\n", out_file)
  close.connection(out_file)
}

# Custom presentation functions ####

# 3D plot optimised for my carapace data
plot_carapace <- function(A, what = NULL, title=FALSE) {
  if (is.character(what)) what <- which(dimnames(A)[[3]] == what)
  bone_links <- c(1, 2, 3, 42, 41, 40, 1, NA, 4, 43, NA, 5, 44, NA, 6, 45, NA, 7, 46, NA, 8, 47, NA, 2, 19, 20, 21, 22, 23, 24, 25, 9, NA, 12, 19, NA, 13, 20, NA, 14, 21, NA, 15, 22, NA, 16, 23, NA, 17, 24, NA, 18, 25, NA, 9, 10, 11, 50, 49, 48, NA, 49, 10, NA, 2, 26, 27, 28, 29, 30, 31, 32, NA, 26, 33, NA, 27, 34, NA, 28, 35, NA, 29, 36, NA, 30, 37, NA, 31, 38, NA, 32, 39, NA, 1, 33, 34, 35, 36, 37, 38, 39, NA, 51, 58, NA, 52, 59, NA, 53, 60, NA, 54, 61, NA, 55, 62, NA, 56, 63, NA, 57, 64, NA, 41, 58, 59, 60, 61, 62, 63, 64, 48, NA, 65, 72, NA, 66, 73, NA, 67, 74, NA, 68, 75, NA, 69, 76, NA, 70, 77, NA, 71, 78, NA, 40, 72, 73, 74, 75, 76, 77, 78, NA, 41, 65, 66, 67, 68, 69, 70, 71, NA, 39, 11, NA, 78, 50, NA, 3, 4, 5, 6, 7, 8, NA, 42, 43, 44, 45, 46, 47, NA, 3, 12:18, NA, 42, 51:57)
  scute_links <- c(79,80,86,81,87,82,88,83,89,84,90,85,NA,86,91,92,93,90,NA,87,91,NA,88,92,NA,89,93,NA,80,94,95,96,97,98,99,100,101,102,103,85,NA,79,104,105,106,107,108,109,110,111,112,113,114,NA,94,104,NA,95,105,NA,96,106,NA,97,107,NA,98,108,NA,99,109,NA,100,110,NA,101,111,NA,102,112,NA,103,113,NA,85,114,NA,115,116,122,117,123,118,124,119,125,120,126,121,NA,115,79,NA,116,80,NA,117,81,NA,118,82,NA,119,83,NA,120,84,NA,122,127,128,129,126,121,NA,123,127,NA,124,128,NA,125,129,NA,116,130,131,132,133,134,135,136,137,138,139,121,NA,115,140,141,142,143,144,145,146,147,148,149,150,NA,130,140,NA,131,141,NA,132,142,NA,133,143,NA,134,144,NA,135,145,NA,136,146,NA,137,147,NA,138,148,NA,139,149,NA,121,150,NA,121,151,85,NA,150,152,114,NA,151,152) - 78
  bones <- 1:78
  scutes <- 79:152
  if (length(dim(A)) == 3) {
    if (title) title <- dimnames(A)[[3]][what]
    else title <- ""
    see3d(A[bones, , what], links=bone_links, pcol="blue", lcol="blue", main=title, new_plot=T)
    see3d(A[scutes, , what], links=scute_links, pcol="red", lcol="red")
  } else {
    see3d(A[bones, ], links=bone_links, pcol="blue", lcol="blue", main="", new_plot=T)
    see3d(A[scutes, ], links=scute_links, pcol="red", lcol="red")
  }
}

# 3D plot optimised for my plastron data
plot_plastron <- function(A, what = NULL, title=FALSE) {
  if (is.character(what)) what <- which(dimnames(A)[[3]] == what)
  bone_links <- c(1,9,10,11,NA,12,13,14,15,2,28,27,26,25,24,23,22,1,NA,1,3,8,4,5,NA,6,7,2,NA,8,9,NA,5,11,NA,6,12,NA,7,14,NA,16,21,17,18,24,NA,21,22,NA,25,19,20,2,NA,20,27,NA,NA,NA,NA)
  scute_links <- c(1,34,35,36,37,38,15,2,28,48,47,46,45,44,1,NA,29,34,NA,39,44,NA,30,35,NA,40,45,NA,31,36,NA,41,46,NA,32,37,NA,42,47,NA,33,38,NA,43,48,NA,1,29,30,31,32,33,2,NA,1,39,40,41,42,43,2)
  bones <- 1:28
  scutes <- 29:52
  if (length(dim(A)) == 3) {
    if (title) title <- dimnames(A)[[3]][what]
    else title <- ""
    see3d(A[bones, , what], links=bone_links, pcol="blue", lcol="blue", main=title, new_plot=T)
    see3d(A[, , what], links=scute_links, pcol="blue", lcol="red")
    see3d(A[scutes, , what], links=NULL, pcol="red", lcol="red")
  } else {
    plot(A[bones, ], asp = 1, )
    see3d(A[bones, ], links=bone_links, pcol="blue", lcol="blue", main="", new_plot=T)
    see3d(A[1:52, ], links=scute_links, pcol="blue", lcol="red")
    see3d(A[scutes, ], links=NULL, pcol="red", lcol="red")
  }
}

# 3D plot of a configuration with links
see3d <- function (mat, links, pcol="black", lcol="black", main="", new_plot=F, ...) {
  link_matrix <- mat[links, ]
  asp <- apply(mat, 2, max) - apply(mat, 2, min)
  asp <- asp / max(asp)
  plot3d(mat, aspect=asp, box=F, axes=F, xlab="", ylab="", zlab="", type ="p", add=(! new_plot), col=pcol, main=main, ...)
  if (! is.null(links)) plot3d(link_matrix, aspect=asp, box=F, axes=F, xlab="", ylab="", zlab="", type ="l", add=TRUE, col=lcol, ...)
}

# Miscellaneous functions ####

# Prune down trees to match their tips, calculate the RF distance, and scale it by the maximum value
scaled_RF <- function(tr1, tr2) {
  common_tips <- intersect(tr1$tip.label, tr2$tip.label)
  tr1 <- drop.tip(tr1, which(tr1$tip.label %notin% common_tips))
  tr2 <- drop.tip(tr2, which(tr2$tip.label %notin% common_tips))
  dist.topo(tr1, tr2) / (2 * (length(common_tips) - 3))
}

# Calculates the matrix of scaled parwise Robinson-Foulds distances from a list of trees
tree_dist_mat <- function (trees) {
  n_trees <- length(trees)
  M <- matrix(NA, ncol = n_trees, nrow = n_trees)
  colnames(M) <- names(trees)
  rownames(M) <- names(trees)
  
  for (i in 1:(n_trees - 1)) {
    for (j in (i + 1):n_trees) {
      M[j, i] <- scaled_RF(trees[[i]], trees[[j]])
    }
  }
  diag(M) <- 0
  return(M)
}


# The following not used in the present script.

ild2 <- function (M1, M2) {
  sqrt(apply((M1 - M2) ^ 2, 1, sum))
}

euclid <- function(dat) {
  p <- dim(dat)[1]
  k <- dim(dat)[2]
  n <- dim(dat)[3]
  n_names <- dimnames(dat)[[3]]
  
  d_mat <- matrix(NA, ncol=n, nrow=n)
  rownames(d_mat) <- n_names
  colnames(d_mat) <- n_names
  
  for (i in 1:n) {
    for (j in i:n) {
      if (i == j) d_mat[i, j] <- 0
      else  d_mat[i, j] <- sqrt(sum(ild2(dat[, , i], dat[, , j])^2))
    }
  }
  return(d_mat)
}

# Write a phylo or multiPhylo object in the TNT parenthetica tree format.
# TNT cannot read trees without a matrix in memory, so this function offers the option to create a dummy matrix.
# Trees can have diffent tip labels.
write.tree.tnt <- function(tr, out_file, dummy = TRUE) {
  require(ape)
  out_file <- file(out_file, "w")
  if (class(tr) == "multiPhylo") {
    n_tr <- length(tr)
    tip_labels <- Reduce(union, lapply(tr, function(x)
      x$tip.label))
  } else {
    n_tr <- 1
    tip_labels <- tr$tip.label
    tr <- list(tr)
  }
  
  if (dummy) {
    writeLines("NSTATES 8 ;", out_file)
    writeLines("XREAD", out_file)
    writeLines(paste("1", length(tip_labels)), out_file)
    mat <- cbind(tip_labels, 0)
    write.table(
      mat,
      file = out_file,
      col.names = FALSE,
      row.names = FALSE,
      append = TRUE,
      quote = FALSE
    )
    writeLines(";", out_file)
  }
  
  tr <-
    lapply(tr, function (x) {
      x$edge.length <- NULL
      x$node.label <- NULL
      x
    })
  
  writeLines("TREAD", out_file)
  for (i in 1:n_tr) {
    nwkstr <- write.tree(tr[[i]])
    tntstr <- gsub(",", " ", nwkstr)
    tntstr <- gsub(";", "*", tntstr)
    writeLines(tntstr, out_file)
  }
  writeLines(";\nPROC/;\n", out_file)
  close.connection(out_file)
}

incomp.bp <- function(ref, trees) {
  require(phangorn)
  if (class(trees) == "phylo") trees <- c(trees)
  ref <- unroot(ref)
  trees <- unroot(trees)
  ref$edge.length <- ref$node.label <- NULL
  bpspace <- as.splits(c(ref, trees))
  bpref <- data.frame(t(as.matrix(as.splits(ref))))
  bptr <- lapply(trees, as.splits)
  bptr <- lapply(lapply(lapply(bptr, function (x) as.matrix(x)[, ref$tip.label]), t), data.frame)
  cmat <- as.matrix(compatible(bpspace))
  bpspace <- data.frame(t(as.matrix(bpspace)))
  bpref <- na.omit(match(bpref, bpspace))
  bptr <- lapply(bptr, function (x) na.omit(match(x, bpspace)))
  sapply(bptr, function(x) sum(rowSums(cmat[x, bpref]) > 0))
}

shared.bp <- function(ref, trees, normalise = TRUE) {
  require(phangorn)
  n_tips <- length(ref$tip.label)
  if (class(trees) == "phylo") trees <- c(trees)
  bpspace <- as.splits(c(ref, trees))
  bpref <- data.frame(t(as.matrix(as.splits(ref))))
  bptr <- lapply(trees, as.splits)
  bptr <- lapply(lapply(lapply(bptr, function (x) as.matrix(x)[, ref$tip.label]), t), data.frame)
  bpspace <- data.frame(t(as.matrix(bpspace)))
  bpref <- match(bpref[-(1:(n_tips + 1))], bpspace)
  bptr <- sapply(bptr, function (x) match(x, bpspace))
  cbp <- sapply(bptr, function (x) sum(x %in% bpref))
  if (normalise) return(cbp/(nrow(ref$edge) - n_tips))
  else return(cbp)
}